#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2018 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""The ahi_hsd reader tests package."""
from __future__ import annotations

import contextlib
import unittest
import warnings
from datetime import datetime
from typing import Any, Dict
from unittest import mock

import dask.array as da
import numpy as np
import pytest

from satpy.readers.ahi_hsd import AHIHSDFileHandler
from satpy.readers.utils import get_geostationary_mask
from satpy.tests.utils import make_dataid

InfoDict = Dict[str, Any]

FAKE_BASIC_INFO: InfoDict = {
    'blocklength': 0,
    'satellite': 'Himawari-8',
    'observation_area': 'FLDK',
    'observation_start_time': 58413.12523839,
    'observation_end_time': 58413.12562439,
    'observation_timeline': '0300',
}
FAKE_DATA_INFO: InfoDict = {
    'blocklength': 50,
    'compression_flag_for_data': 0,
    'hblock_number': 2,
    'number_of_bits_per_pixel': 16,
    'number_of_columns': 11000,
    'number_of_lines': 1100,
    'spare': '',
}
FAKE_PROJ_INFO: InfoDict = {
    'CFAC': 40932549,
    'COFF': 5500.5,
    'LFAC': 40932549,
    'LOFF': 5500.5,
    'blocklength': 127,
    'coeff_for_sd': 1737122264.0,
    'distance_from_earth_center': 42164.0,
    'earth_equatorial_radius': 6378.137,
    'earth_polar_radius': 6356.7523,
    'hblock_number': 3,
    'req2_rpol2': 1.006739501,
    'req2_rpol2_req2': 0.0066943844,
    'resampling_size': 4,
    'resampling_types': 0,
    'rpol2_req2': 0.993305616,
    'spare': '',
    'sub_lon': 140.7,
}
FAKE_NAV_INFO: InfoDict = {
    'SSP_longitude': 140.65699999999998,
    'SSP_latitude': 0.0042985719753897015,
    'distance_earth_center_to_satellite': 42165.04,
    'nadir_longitude': 140.25253875463318,
    'nadir_latitude': 0.01674775121155575,
}
FAKE_CAL_INFO: InfoDict = {'blocklength': 0, 'band_number': [4]}
FAKE_IRVISCAL_INFO: InfoDict = {}
FAKE_INTERCAL_INFO: InfoDict = {'blocklength': 0}
FAKE_SEGMENT_INFO: InfoDict = {'blocklength': 0}
FAKE_NAVCORR_INFO: InfoDict = {'blocklength': 0, 'numof_correction_info_data': [1]}
FAKE_NAVCORR_SUBINFO: InfoDict = {}
FAKE_OBS_TIME_INFO: InfoDict = {'blocklength': 0, 'number_of_observation_times': [1]}
FAKE_OBS_LINETIME_INFO: InfoDict = {}
FAKE_ERROR_INFO: InfoDict = {'blocklength': 0, 'number_of_error_info_data': [1]}
FAKE_ERROR_LINE_INFO: InfoDict = {}
FAKE_SPARE_INFO: InfoDict = {'blocklength': 0}


def _new_unzip(fname, prefix=''):
    """Fake unzipping."""
    if fname[-3:] == 'bz2':
        return prefix + fname[:-4]
    return fname


class TestAHIHSDNavigation(unittest.TestCase):
    """Test the AHI HSD reader navigation."""

    @mock.patch('satpy.readers.ahi_hsd.np2str')
    @mock.patch('satpy.readers.ahi_hsd.np.fromfile')
    def test_region(self, fromfile, np2str):
        """Test region navigation."""
        from pyresample.utils import proj4_radius_parameters
        np2str.side_effect = lambda x: x
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_hsd.open', m, create=True):
            fh = AHIHSDFileHandler('somefile',
                                   {'segment': 1, 'total_segments': 1},
                                   filetype_info={'file_type': 'hsd_b01'},
                                   user_calibration=None)
            fh.proj_info = {'CFAC': 40932549,
                            'COFF': -591.5,
                            'LFAC': 40932549,
                            'LOFF': 5132.5,
                            'blocklength': 127,
                            'coeff_for_sd': 1737122264.0,
                            'distance_from_earth_center': 42164.0,
                            'earth_equatorial_radius': 6378.137,
                            'earth_polar_radius': 6356.7523,
                            'hblock_number': 3,
                            'req2_rpol2': 1.006739501,
                            'req2_rpol2_req2': 0.0066943844,
                            'resampling_size': 4,
                            'resampling_types': 0,
                            'rpol2_req2': 0.993305616,
                            'spare': '',
                            'sub_lon': 140.7}

            fh.data_info = {'blocklength': 50,
                            'compression_flag_for_data': 0,
                            'hblock_number': 2,
                            'number_of_bits_per_pixel': 16,
                            'number_of_columns': 1000,
                            'number_of_lines': 1000,
                            'spare': ''}

            area_def = fh.get_area_def(None)
            proj_dict = area_def.proj_dict
            a, b = proj4_radius_parameters(proj_dict)
            self.assertEqual(a, 6378137.0)
            self.assertEqual(b, 6356752.3)
            self.assertEqual(proj_dict['h'], 35785863.0)
            self.assertEqual(proj_dict['lon_0'], 140.7)
            self.assertEqual(proj_dict['proj'], 'geos')
            self.assertEqual(proj_dict['units'], 'm')
            np.testing.assert_allclose(area_def.area_extent, (592000.0038256242, 4132000.0267018233,
                                                              1592000.0102878273, 5132000.033164027))

    @mock.patch('satpy.readers.ahi_hsd.np2str')
    @mock.patch('satpy.readers.ahi_hsd.np.fromfile')
    def test_segment(self, fromfile, np2str):
        """Test segment navigation."""
        from pyresample.utils import proj4_radius_parameters
        np2str.side_effect = lambda x: x
        m = mock.mock_open()
        with mock.patch('satpy.readers.ahi_hsd.open', m, create=True):
            fh = AHIHSDFileHandler('somefile', {'segment': 8, 'total_segments': 10},
                                   filetype_info={'file_type': 'hsd_b01'})
            fh.proj_info = {'CFAC': 40932549,
                            'COFF': 5500.5,
                            'LFAC': 40932549,
                            'LOFF': 5500.5,
                            'blocklength': 127,
                            'coeff_for_sd': 1737122264.0,
                            'distance_from_earth_center': 42164.0,
                            'earth_equatorial_radius': 6378.137,
                            'earth_polar_radius': 6356.7523,
                            'hblock_number': 3,
                            'req2_rpol2': 1.006739501,
                            'req2_rpol2_req2': 0.0066943844,
                            'resampling_size': 4,
                            'resampling_types': 0,
                            'rpol2_req2': 0.993305616,
                            'spare': '',
                            'sub_lon': 140.7}

            fh.data_info = {'blocklength': 50,
                            'compression_flag_for_data': 0,
                            'hblock_number': 2,
                            'number_of_bits_per_pixel': 16,
                            'number_of_columns': 11000,
                            'number_of_lines': 1100,
                            'spare': ''}

            area_def = fh.get_area_def(None)
            proj_dict = area_def.proj_dict
            a, b = proj4_radius_parameters(proj_dict)
            self.assertEqual(a, 6378137.0)
            self.assertEqual(b, 6356752.3)
            self.assertEqual(proj_dict['h'], 35785863.0)
            self.assertEqual(proj_dict['lon_0'], 140.7)
            self.assertEqual(proj_dict['proj'], 'geos')
            self.assertEqual(proj_dict['units'], 'm')
            np.testing.assert_allclose(area_def.area_extent, (-5500000.035542117, -3300000.021325271,
                                                              5500000.035542117, -2200000.0142168473))


class TestAHIHSDFileHandler:
    """Tests for the AHI HSD file handler."""

    def test_bad_calibration(self):
        """Test that a bad calibration mode causes an exception."""
        with pytest.raises(ValueError):
            with _fake_hsd_handler(fh_kwargs={"calib_mode": "BAD_MODE"}):
                pass

    @pytest.mark.parametrize(
        ("round_actual_position", "expected_result"),
        [
            (False, (140.65699999999998, 0.0042985719753897015, 35786903.00011936)),
            (True, (140.657, 0.0, 35786850.0))
        ]
    )
    def test_actual_satellite_position(self, round_actual_position, expected_result):
        """Test that rounding of the actual satellite position can be controlled."""
        with _fake_hsd_handler(fh_kwargs={"round_actual_position": round_actual_position}) as fh:
            ds_id = make_dataid(name="B01")
            ds_info = {
                "units": "%",
                "standard_name": "some_name",
                "wavelength": (0.1, 0.2, 0.3),
            }
            metadata = fh._get_metadata(ds_id, ds_info)
            orb_params = metadata["orbital_parameters"]
            assert orb_params["satellite_actual_longitude"] == expected_result[0]
            assert orb_params["satellite_actual_latitude"] == expected_result[1]
            assert orb_params["satellite_actual_altitude"] == expected_result[2]

    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._check_fpos')
    def test_read_header(self, *mocks):
        """Test header reading."""
        with _fake_hsd_handler() as fh:
            fh._read_header(mock.MagicMock())

    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._read_data')
    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._mask_invalid')
    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler.calibrate')
    def test_read_band(self, calibrate, *mocks):
        """Test masking of space pixels."""
        nrows = 25
        ncols = 100
        calibrate.return_value = np.ones((nrows, ncols))
        with _fake_hsd_handler() as fh:
            fh.data_info['number_of_columns'] = ncols
            fh.data_info['number_of_lines'] = nrows
            im = fh.read_band(mock.MagicMock(), mock.MagicMock())
            # Note: Within the earth's shape get_geostationary_mask() is True but the numpy.ma mask
            # is False
            mask = im.to_masked_array().mask
            ref_mask = np.logical_not(get_geostationary_mask(fh.area).compute())
            np.testing.assert_equal(mask, ref_mask)

            # Test attributes
            orb_params_exp = {'projection_longitude': 140.7,
                              'projection_latitude': 0.,
                              'projection_altitude': 35785863.0,
                              'satellite_actual_longitude': 140.657,
                              'satellite_actual_latitude': 0.0,
                              'satellite_actual_altitude': 35786850,
                              'nadir_longitude': 140.252539,
                              'nadir_latitude': 0.01674775}
            actual_obs_params = im.attrs['orbital_parameters']
            for key, value in orb_params_exp.items():
                assert key in actual_obs_params
                np.testing.assert_allclose(value, actual_obs_params[key])

            time_params_exp = {
                'nominal_start_time': datetime(2018, 10, 22, 3, 0, 0, 0),
                'nominal_end_time': datetime(2018, 10, 22, 3, 0, 0, 0),
                'observation_start_time': datetime(2018, 10, 22, 3, 0, 20, 596896),
                'observation_end_time': datetime(2018, 10, 22, 3, 0, 53, 947296),
            }
            actual_time_params = im.attrs['time_parameters']
            for key, value in time_params_exp.items():
                assert key in actual_time_params
                assert value == actual_time_params[key]

            # Test if masking space pixels disables with appropriate flag
            fh.mask_space = False
            with mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._mask_space') as mask_space:
                fh.read_band(mock.MagicMock(), mock.MagicMock())
                mask_space.assert_not_called()

    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._read_data')
    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._mask_invalid')
    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler.calibrate')
    def test_scene_loading(self, calibrate, *mocks):
        """Test masking of space pixels."""
        from satpy import Scene
        nrows = 25
        ncols = 100
        calibrate.return_value = np.ones((nrows, ncols))
        with _fake_hsd_handler() as fh:
            with mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler') as fh_cls:
                fh_cls.return_value = fh
                fh.filename_info['total_segments'] = 1
                fh.filename_info['segment'] = 1
                fh.data_info['number_of_columns'] = ncols
                fh.data_info['number_of_lines'] = nrows
                scn = Scene(reader='ahi_hsd', filenames=['HS_H08_20210225_0700_B07_FLDK_R20_S0110.DAT'])
                scn.load(['B07'])
                im = scn['B07']

                # Make sure space masking worked
                mask = im.to_masked_array().mask
                ref_mask = np.logical_not(get_geostationary_mask(fh.area).compute())
                np.testing.assert_equal(mask, ref_mask)

    def test_time_properties(self):
        """Test start/end/scheduled time properties."""
        with _fake_hsd_handler() as fh:
            assert fh.start_time == datetime(2018, 10, 22, 3, 0)
            assert fh.end_time == datetime(2018, 10, 22, 3, 0)
            assert fh.observation_start_time == datetime(2018, 10, 22, 3, 0, 20, 596896)
            assert fh.observation_end_time == datetime(2018, 10, 22, 3, 0, 53, 947296)
            assert fh.nominal_start_time == datetime(2018, 10, 22, 3, 0, 0, 0)
            assert fh.nominal_end_time == datetime(2018, 10, 22, 3, 0, 0, 0)

    def test_scanning_frequencies(self):
        """Test scanning frequencies."""
        with _fake_hsd_handler() as fh:
            fh.observation_area = 'JP04'
            assert fh.nominal_start_time == datetime(2018, 10, 22, 3, 7, 30, 0)
            assert fh.nominal_end_time == datetime(2018, 10, 22, 3, 7, 30, 0)
            fh.observation_area = 'R304'
            assert fh.nominal_start_time == datetime(2018, 10, 22, 3, 7, 30, 0)
            assert fh.nominal_end_time == datetime(2018, 10, 22, 3, 7, 30, 0)
            fh.observation_area = 'R420'
            assert fh.nominal_start_time == datetime(2018, 10, 22, 3, 9, 30, 0)
            assert fh.nominal_end_time == datetime(2018, 10, 22, 3, 9, 30, 0)
            fh.observation_area = 'R520'
            assert fh.nominal_start_time == datetime(2018, 10, 22, 3, 9, 30, 0)
            assert fh.nominal_end_time == datetime(2018, 10, 22, 3, 9, 30, 0)
            fh.observation_area = 'FLDK'
            assert fh.nominal_start_time == datetime(2018, 10, 22, 3, 0, 0, 0)
            assert fh.nominal_end_time == datetime(2018, 10, 22, 3, 0, 0, 0)

    def test_blocklen_error(self, *mocks):
        """Test erraneous blocklength."""
        open_name = '%s.open' % __name__
        fpos = 50
        with _fake_hsd_handler() as fh, \
                mock.patch(open_name, create=True) as mock_open, \
                mock_open(mock.MagicMock(), 'r') as fp_:
            # Expected and actual blocklength match
            fp_.tell.return_value = 50
            with warnings.catch_warnings(record=True) as w:
                fh._check_fpos(fp_, fpos, 0, 'header 1')
                assert len(w) == 0

            # Expected and actual blocklength do not match
            fp_.tell.return_value = 100
            with warnings.catch_warnings(record=True) as w:
                fh._check_fpos(fp_, fpos, 0, 'header 1')
                assert len(w) > 0

    def test_is_valid_time(self):
        """Test that valid times are correctly identified."""
        assert AHIHSDFileHandler._is_valid_timeline(FAKE_BASIC_INFO['observation_timeline'])
        assert not AHIHSDFileHandler._is_valid_timeline('65526')

    def test_time_rounding(self):
        """Test rounding of the nominal time."""
        mocker = mock.MagicMock()
        in_date = datetime(2020, 1, 1, 12, 0, 0)

        with mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler._is_valid_timeline', mocker):
            with _fake_hsd_handler() as fh:
                mocker.return_value = True
                assert fh._modify_observation_time_for_nominal(in_date) == datetime(2020, 1, 1, 3, 0, 0)
                mocker.return_value = False
                assert fh._modify_observation_time_for_nominal(in_date) == datetime(2020, 1, 1, 12, 0, 0)


class TestAHICalibration(unittest.TestCase):
    """Test case for various AHI calibration types."""

    @mock.patch('satpy.readers.ahi_hsd.AHIHSDFileHandler.__init__',
                return_value=None)
    def setUp(self, *mocks):
        """Create fake data for testing."""
        self.def_cali = [-0.0037, 15.20]
        self.upd_cali = [-0.0074, 30.40]
        self.bad_cali = [0.0, 0.0]
        fh = AHIHSDFileHandler(filetype_info={'file_type': 'hsd_b01'})
        fh.calib_mode = 'NOMINAL'
        fh.user_calibration = None
        fh.is_zipped = False
        fh._header = {
            'block5': {'band_number': [5],
                       'gain_count2rad_conversion': [self.def_cali[0]],
                       'offset_count2rad_conversion': [self.def_cali[1]],
                       'central_wave_length': [10.4073], },
            'calibration': {'coeff_rad2albedo_conversion': [0.0019255],
                            'speed_of_light': [299792458.0],
                            'planck_constant': [6.62606957e-34],
                            'boltzmann_constant': [1.3806488e-23],
                            'c0_rad2tb_conversion': [-0.116127314574],
                            'c1_rad2tb_conversion': [1.00099153832],
                            'c2_rad2tb_conversion': [-1.76961091571e-06],
                            'cali_gain_count2rad_conversion': [self.upd_cali[0]],
                            'cali_offset_count2rad_conversion': [self.upd_cali[1]]},
        }

        self.counts = da.array(np.array([[0., 1000.],
                                         [2000., 5000.]]))
        self.fh = fh

    def test_default_calibrate(self, *mocks):
        """Test default in-file calibration modes."""
        self.setUp()
        # Counts
        self.assertEqual(self.fh.calibrate(data=123,
                                           calibration='counts'),
                         123)

        # Radiance
        rad_exp = np.array([[15.2, 11.5],
                            [7.8, -3.3]])
        rad = self.fh.calibrate(data=self.counts,
                                calibration='radiance')
        self.assertTrue(np.allclose(rad, rad_exp))

        # Brightness Temperature
        bt_exp = np.array([[330.978979, 310.524688],
                           [285.845017, np.nan]])
        bt = self.fh.calibrate(data=self.counts,
                               calibration='brightness_temperature')
        np.testing.assert_allclose(bt, bt_exp)

        # Reflectance
        refl_exp = np.array([[2.92676, 2.214325],
                             [1.50189, 0.]])
        refl = self.fh.calibrate(data=self.counts,
                                 calibration='reflectance')
        self.assertTrue(np.allclose(refl, refl_exp))

    def test_updated_calibrate(self):
        """Test updated in-file calibration modes."""
        # Standard operation
        self.fh.calib_mode = 'UPDATE'
        rad_exp = np.array([[30.4, 23.0],
                            [15.6, -6.6]])
        rad = self.fh.calibrate(data=self.counts, calibration='radiance')
        self.assertTrue(np.allclose(rad, rad_exp))

        # Case for no updated calibration available (older data)
        self.fh._header = {
            'block5': {'band_number': [5],
                       'gain_count2rad_conversion': [self.def_cali[0]],
                       'offset_count2rad_conversion': [self.def_cali[1]],
                       'central_wave_length': [10.4073], },
            'calibration': {'coeff_rad2albedo_conversion': [0.0019255],
                            'speed_of_light': [299792458.0],
                            'planck_constant': [6.62606957e-34],
                            'boltzmann_constant': [1.3806488e-23],
                            'c0_rad2tb_conversion': [-0.116127314574],
                            'c1_rad2tb_conversion': [1.00099153832],
                            'c2_rad2tb_conversion': [-1.76961091571e-06],
                            'cali_gain_count2rad_conversion': [self.bad_cali[0]],
                            'cali_offset_count2rad_conversion': [self.bad_cali[1]]},
        }
        rad = self.fh.calibrate(data=self.counts, calibration='radiance')
        rad_exp = np.array([[15.2, 11.5],
                            [7.8, -3.3]])
        self.assertTrue(np.allclose(rad, rad_exp))

    def test_user_calibration(self):
        """Test user-defined calibration modes."""
        # This is for radiance correction
        self.fh.user_calibration = {'B13': {'slope': 0.95,
                                            'offset': -0.1}}
        self.fh.band_name = 'B13'
        rad = self.fh.calibrate(data=self.counts, calibration='radiance').compute()
        rad_exp = np.array([[16.10526316, 12.21052632],
                            [8.31578947, -3.36842105]])
        self.assertTrue(np.allclose(rad, rad_exp))

        # This is for DN calibration
        self.fh.user_calibration = {'B13': {'slope': -0.0032,
                                            'offset': 15.20},
                                    'type': 'DN'}
        self.fh.band_name = 'B13'
        rad = self.fh.calibrate(data=self.counts, calibration='radiance').compute()
        rad_exp = np.array([[15.2, 12.],
                            [8.8, -0.8]])
        self.assertTrue(np.allclose(rad, rad_exp))


@contextlib.contextmanager
def _fake_hsd_handler(fh_kwargs=None):
    """Create a test file handler."""
    m = mock.mock_open()
    with mock.patch('satpy.readers.ahi_hsd.np.fromfile', _custom_fromfile), \
            mock.patch('satpy.readers.ahi_hsd.unzip_file', mock.MagicMock(side_effect=_new_unzip)), \
            mock.patch('satpy.readers.ahi_hsd.open', m, create=True):
        in_fname = 'test_file.bz2'
        fh = _create_fake_file_handler(in_fname, fh_kwargs=fh_kwargs)
        yield fh


def _custom_fromfile(*args, **kwargs):
    from satpy.readers.ahi_hsd import (
        _BASIC_INFO_TYPE,
        _CAL_INFO_TYPE,
        _DATA_INFO_TYPE,
        _ERROR_INFO_TYPE,
        _ERROR_LINE_INFO_TYPE,
        _INTER_CALIBRATION_INFO_TYPE,
        _IRCAL_INFO_TYPE,
        _NAV_INFO_TYPE,
        _NAVIGATION_CORRECTION_INFO_TYPE,
        _NAVIGATION_CORRECTION_SUBINFO_TYPE,
        _OBSERVATION_LINE_TIME_INFO_TYPE,
        _OBSERVATION_TIME_INFO_TYPE,
        _PROJ_INFO_TYPE,
        _SEGMENT_INFO_TYPE,
        _SPARE_TYPE,
        _VISCAL_INFO_TYPE,
    )
    dtype = kwargs.get("dtype")
    fake_info_map = {
        _BASIC_INFO_TYPE: FAKE_BASIC_INFO,
        _DATA_INFO_TYPE: FAKE_DATA_INFO,
        _NAV_INFO_TYPE: FAKE_NAV_INFO,
        _PROJ_INFO_TYPE: FAKE_PROJ_INFO,
        _CAL_INFO_TYPE: FAKE_CAL_INFO,
        _VISCAL_INFO_TYPE: FAKE_IRVISCAL_INFO,
        _IRCAL_INFO_TYPE: FAKE_IRVISCAL_INFO,
        _INTER_CALIBRATION_INFO_TYPE: FAKE_INTERCAL_INFO,
        _SEGMENT_INFO_TYPE: FAKE_SEGMENT_INFO,
        _NAVIGATION_CORRECTION_INFO_TYPE: FAKE_NAVCORR_INFO,
        _NAVIGATION_CORRECTION_SUBINFO_TYPE: FAKE_NAVCORR_SUBINFO,
        _OBSERVATION_TIME_INFO_TYPE: FAKE_OBS_TIME_INFO,
        _OBSERVATION_LINE_TIME_INFO_TYPE: FAKE_OBS_LINETIME_INFO,
        _ERROR_INFO_TYPE: FAKE_ERROR_INFO,
        _ERROR_LINE_INFO_TYPE: FAKE_ERROR_LINE_INFO,
        _SPARE_TYPE: FAKE_SPARE_INFO,
    }
    info_dict = fake_info_map[dtype]
    fake_arr = np.zeros((1,), dtype=dtype)
    for key, value in info_dict.items():
        fake_arr[key] = value
    return fake_arr


def _create_fake_file_handler(in_fname, filename_info=None, filetype_info=None, fh_kwargs=None):
    if filename_info is None:
        filename_info = {'segment': 8, 'total_segments': 10}
    if filetype_info is None:
        filetype_info = {'file_type': 'hsd_b01'}
    if fh_kwargs is None:
        fh_kwargs = {}
    fh = AHIHSDFileHandler(in_fname, filename_info, filetype_info, **fh_kwargs)

    # Check that the filename is altered and 2 digit segment prefix added for bz2 format files
    assert in_fname != fh.filename
    assert str(filename_info['segment']).zfill(2) == fh.filename[0:2]
    return fh



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009-2019 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Pygac interface."""

from datetime import datetime
from unittest import TestCase, mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

GAC_PATTERN = '{creation_site:3s}.{transfer_mode:4s}.{platform_id:2s}.D{start_time:%y%j.S%H%M}.E{end_time:%H%M}.B{orbit_number:05d}{end_orbit_last_digits:02d}.{station:2s}'  # noqa

GAC_POD_FILENAMES = ['NSS.GHRR.NA.D79184.S1150.E1337.B0008384.WI',
                     'NSS.GHRR.NA.D79184.S2350.E0137.B0008384.WI',
                     'NSS.GHRR.NA.D80021.S0927.E1121.B0295354.WI',
                     'NSS.GHRR.NA.D80021.S1120.E1301.B0295455.WI',
                     'NSS.GHRR.NA.D80021.S1256.E1450.B0295556.GC',
                     'NSS.GHRR.NE.D83208.S1219.E1404.B0171819.WI',
                     'NSS.GHRR.NG.D88002.S0614.E0807.B0670506.WI',
                     'NSS.GHRR.TN.D79183.S1258.E1444.B0369697.GC',
                     'NSS.GHRR.TN.D80003.S1147.E1332.B0630506.GC',
                     'NSS.GHRR.TN.D80003.S1328.E1513.B0630507.GC',
                     'NSS.GHRR.TN.D80003.S1509.E1654.B0630608.GC']

GAC_KLM_FILENAMES = ['NSS.GHRR.NK.D01235.S0252.E0446.B1703233.GC',
                     'NSS.GHRR.NL.D01288.S2315.E0104.B0549495.GC',
                     'NSS.GHRR.NM.D04111.S2305.E0050.B0947778.GC',
                     'NSS.GHRR.NN.D13011.S0559.E0741.B3939192.WI',
                     'NSS.GHRR.NP.D15361.S0121.E0315.B3547172.SV',
                     'NSS.GHRR.M1.D15362.S0031.E0129.B1699697.SV',
                     'NSS.GHRR.M2.D10178.S2359.E0142.B1914142.SV']

LAC_POD_FILENAMES = ['BRN.HRPT.ND.D95152.S1730.E1715.B2102323.UB',
                     'BRN.HRPT.ND.D95152.S1910.E1857.B2102424.UB',
                     'BRN.HRPT.NF.D85152.S1345.E1330.B0241414.UB',
                     'BRN.HRPT.NJ.D95152.S1233.E1217.B0216060.UB']

LAC_KLM_FILENAMES = ['BRN.HRPT.M1.D14152.S0958.E1012.B0883232.UB',
                     'BRN.HRPT.M1.D14152.S1943.E1958.B0883838.UB',
                     'BRN.HRPT.M2.D12153.S0912.E0922.B2914747.UB',
                     'BRN.HRPT.NN.D12153.S0138.E0152.B3622828.UB',
                     'BRN.HRPT.NN.D12153.S0139.E0153.B3622828.UB',
                     'BRN.HRPT.NN.D12153.S1309.E1324.B3623535.UB',
                     'BRN.HRPT.NP.D12153.S0003.E0016.B1707272.UB',
                     'BRN.HRPT.NP.D12153.S1134.E1148.B1707979.UB',
                     'BRN.HRPT.NP.D16184.S1256.E1311.B3813131.UB',
                     'BRN.HRPT.NP.D16184.S1438.E1451.B3813232.UB',
                     'BRN.HRPT.NP.D16184.S1439.E1451.B3813232.UB',
                     'BRN.HRPT.NP.D16185.S1245.E1259.B3814545.UB',
                     'BRN.HRPT.NP.D16185.S1427.E1440.B3814646.UB',
                     'NSS.FRAC.M2.D12153.S1729.E1910.B2915354.SV',
                     'NSS.LHRR.NP.D16306.S1803.E1814.B3985555.WI']


@mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile.__init__', return_value=None)
def _get_fh_mocked(init_mock, **attrs):
    """Create a mocked file handler with the given attributes."""
    from satpy.readers.avhrr_l1b_gaclac import GACLACFile

    fh = GACLACFile()
    for name, value in attrs.items():
        setattr(fh, name, value)
    return fh


def _get_reader_mocked(along_track=3):
    """Create a mocked reader."""
    reader = mock.MagicMock(spacecraft_name='spacecraft_name',
                            meta_data={'foo': 'bar'})
    reader.mask = [0, 0]
    reader.get_times.return_value = np.arange(along_track)
    reader.get_tle_lines.return_value = 'tle'
    return reader


class PygacPatcher(TestCase):
    """Patch pygac."""

    def setUp(self):
        """Patch pygac imports."""
        self.pygac = mock.MagicMock()
        self.fhs = mock.MagicMock()
        modules = {
            'pygac': self.pygac,
            'pygac.gac_klm': self.pygac.gac_klm,
            'pygac.gac_pod': self.pygac.gac_pod,
            'pygac.lac_klm': self.pygac.lac_klm,
            'pygac.lac_pod': self.pygac.lac_pod,
            'pygac.utils': self.pygac.utils,
            'pygac.calibration': self.pygac.calibration,
        }

        self.module_patcher = mock.patch.dict('sys.modules', modules)
        self.module_patcher.start()

    def tearDown(self):
        """Unpatch the pygac imports."""
        self.module_patcher.stop()


class GACLACFilePatcher(PygacPatcher):
    """Patch pygac."""

    def setUp(self):
        """Patch GACLACFile."""
        super().setUp()

        # Import GACLACFile here to make it patchable. Otherwise self._get_fh
        # might import it first which would prevent a successful patch.
        from satpy.readers.avhrr_l1b_gaclac import GACLACFile
        self.GACLACFile = GACLACFile


class TestGACLACFile(GACLACFilePatcher):
    """Test the GACLAC file handler."""

    def _get_fh(self, filename='NSS.GHRR.NG.D88002.S0614.E0807.B0670506.WI',
                **kwargs):
        """Create a file handler."""
        from trollsift import parse
        filename_info = parse(GAC_PATTERN, filename)
        return self.GACLACFile(filename, filename_info, {}, **kwargs)

    def test_init(self):
        """Test GACLACFile initialization."""
        from pygac.gac_klm import GACKLMReader
        from pygac.gac_pod import GACPODReader
        from pygac.lac_klm import LACKLMReader
        from pygac.lac_pod import LACPODReader

        kwargs = {'start_line': 1,
                  'end_line': 2,
                  'strip_invalid_coords': True,
                  'interpolate_coords': True,
                  'adjust_clock_drift': True,
                  'tle_dir': 'tle_dir',
                  'tle_name': 'tle_name',
                  'tle_thresh': 123,
                  'calibration': 'calibration'}
        for filenames, reader_cls in zip([GAC_POD_FILENAMES, GAC_KLM_FILENAMES, LAC_POD_FILENAMES, LAC_KLM_FILENAMES],
                                         [GACPODReader, GACKLMReader, LACPODReader, LACKLMReader]):
            for filename in filenames:
                fh = self._get_fh(filename, **kwargs)
                self.assertLess(fh.start_time, fh.end_time,
                                "Start time must precede end time.")
                self.assertIs(fh.reader_class, reader_cls,
                              'Wrong reader class assigned to {}'.format(filename))

    def test_read_raw_data(self):
        """Test raw data reading."""
        fh = _get_fh_mocked(reader=None,
                            interpolate_coords='interpolate_coords',
                            creation_site='creation_site',
                            reader_kwargs={'foo': 'bar'},
                            filename='myfile')
        reader = mock.MagicMock(mask=[0])
        reader_cls = mock.MagicMock(return_value=reader)
        fh.reader_class = reader_cls
        fh.read_raw_data()
        reader_cls.assert_called_with(interpolate_coords='interpolate_coords',
                                      creation_site='creation_site',
                                      foo='bar')
        reader.read.assert_called_with('myfile')

        # Test exception if all data is masked
        reader.mask = [1]
        fh.reader = None
        with self.assertRaises(ValueError):
            fh.read_raw_data()

    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile._update_attrs')
    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile.slice')
    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile._get_channel')
    def test_get_dataset_slice(self, get_channel, slc, *mocks):
        """Get a slice of a dataset."""
        from satpy.tests.utils import make_dataid

        # Test slicing/stripping
        def slice_patched(data, times):
            if len(data.shape) == 2:
                return data[1:3, :], times[1:3]
            return data[1:3], times[1:3]

        ch = np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 9],
                       [10, 11, 12],
                       [13, 14, 15]])
        acq = np.array([0, 1, 2, 3, 4])
        slc.side_effect = slice_patched
        get_channel.return_value = ch
        kwargs_list = [{'strip_invalid_coords': False,
                        'start_line': 123, 'end_line': 456},
                       {'strip_invalid_coords': True,
                        'start_line': None, 'end_line': None},
                       {'strip_invalid_coords': True,
                        'start_line': 123, 'end_line': 456}]
        for kwargs in kwargs_list:
            fh = _get_fh_mocked(
                reader=_get_reader_mocked(along_track=len(acq)),
                chn_dict={'1': 0},
                **kwargs
            )

            key = make_dataid(name='1', calibration='reflectance')
            info = {'name': '1', 'standard_name': 'reflectance'}
            res = fh.get_dataset(key, info)
            np.testing.assert_array_equal(res.data, ch[1:3, :])
            np.testing.assert_array_equal(res.coords['acq_time'].data, acq[1:3])
            np.testing.assert_array_equal(slc.call_args_list[-1][1]['times'], acq)
            np.testing.assert_array_equal(slc.call_args_list[-1][1]['data'], ch)

    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile._update_attrs')
    def test_get_dataset_latlon(self, *mocks):
        """Test getting the latitudes and longitudes."""
        from satpy.tests.utils import make_dataid

        lons = np.ones((3, 3))
        lats = 2 * lons
        reader = _get_reader_mocked()
        reader.get_lonlat.return_value = lons, lats
        fh = _get_fh_mocked(
            reader=reader,
            start_line=None,
            end_line=None,
            strip_invalid_coords=False,
            interpolate_coords=True
        )

        # With interpolation of coordinates
        for name, exp_data in zip(['longitude', 'latitude'], [lons, lats]):
            key = make_dataid(name=name)
            info = {'name': name, 'standard_name': 'my_standard_name'}
            res = fh.get_dataset(key=key, info=info)
            exp = xr.DataArray(exp_data,
                               name=res.name,
                               dims=('y', 'x'),
                               coords={'acq_time': ('y', [0, 1, 2])})
            xr.testing.assert_equal(res, exp)

        # Without interpolation of coordinates
        fh.interpolate_coords = False
        for name, _exp_data in zip(['longitude', 'latitude'], [lons, lats]):
            key = make_dataid(name=name)
            info = {'name': name, 'standard_name': 'my_standard_name'}
            res = fh.get_dataset(key=key, info=info)
            self.assertTupleEqual(res.dims, ('y', 'x_every_eighth'))

    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile._update_attrs')
    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile._get_angle')
    def test_get_dataset_angles(self, get_angle, *mocks):
        """Test getting the angles."""
        from satpy.readers.avhrr_l1b_gaclac import ANGLES
        from satpy.tests.utils import make_dataid

        ones = np.ones((3, 3))
        get_angle.return_value = ones
        reader = _get_reader_mocked()
        fh = _get_fh_mocked(
            reader=reader,
            start_line=None,
            end_line=None,
            strip_invalid_coords=False,
            interpolate_coords=True
        )

        # With interpolation of coordinates
        for angle in ANGLES:
            key = make_dataid(name=angle)
            info = {'name': angle, 'standard_name': 'my_standard_name'}
            res = fh.get_dataset(key=key, info=info)
            exp = xr.DataArray(ones,
                               name=res.name,
                               dims=('y', 'x'),
                               coords={'acq_time': ('y', [0, 1, 2])})
            xr.testing.assert_equal(res, exp)

        # Without interpolation of coordinates
        fh.interpolate_coords = False
        for angle in ANGLES:
            key = make_dataid(name=angle)
            info = {'name': angle, 'standard_name': 'my_standard_name'}
            res = fh.get_dataset(key=key, info=info)
            self.assertTupleEqual(res.dims, ('y', 'x_every_eighth'))

    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile._update_attrs')
    def test_get_dataset_qual_flags(self, *mocks):
        """Test getting the qualitiy flags."""
        from satpy.tests.utils import make_dataid

        qual_flags = np.ones((3, 7))
        reader = _get_reader_mocked()
        reader.get_qual_flags.return_value = qual_flags
        fh = _get_fh_mocked(
            reader=reader,
            start_line=None,
            end_line=None,
            strip_invalid_coords=False,
            interpolate_coords=True
        )

        key = make_dataid(name='qual_flags')
        info = {'name': 'qual_flags'}
        res = fh.get_dataset(key=key, info=info)
        exp = xr.DataArray(qual_flags,
                           name=res.name,
                           dims=('y', 'num_flags'),
                           coords={'acq_time': ('y', [0, 1, 2]),
                                   'num_flags': ['Scan line number',
                                                 'Fatal error flag',
                                                 'Insufficient data for calibration',
                                                 'Insufficient data for calibration',
                                                 'Solar contamination of blackbody in channels 3',
                                                 'Solar contamination of blackbody in channels 4',
                                                 'Solar contamination of blackbody in channels 5']})
        xr.testing.assert_equal(res, exp)

    def test_get_channel(self):
        """Test getting the channels."""
        from satpy.tests.utils import make_dataid

        counts = np.moveaxis(np.array([[[1, 2, 3],
                                        [4, 5, 6]]]), 0, 2)
        calib_channels = 2 * counts
        reader = _get_reader_mocked()
        reader.get_counts.return_value = counts
        reader.get_calibrated_channels.return_value = calib_channels
        fh = _get_fh_mocked(reader=reader, counts=None, calib_channels=None,
                            chn_dict={'1': 0})

        key = make_dataid(name='1', calibration='counts')
        # Counts
        res = fh._get_channel(key=key)
        np.testing.assert_array_equal(res, [[1, 2, 3],
                                            [4, 5, 6]])
        np.testing.assert_array_equal(fh.counts, counts)

        # Reflectance and Brightness Temperature
        for calib in ['reflectance', 'brightness_temperature']:
            key = make_dataid(name='1', calibration=calib)
            res = fh._get_channel(key=key)
            np.testing.assert_array_equal(res, [[2, 4, 6],
                                                [8, 10, 12]])
            np.testing.assert_array_equal(fh.calib_channels, calib_channels)

        # Invalid
        with pytest.raises(ValueError):
            key = make_dataid(name='7', calibration='coffee')

        # Buffering
        reader.get_counts.reset_mock()
        key = make_dataid(name='1', calibration='counts')
        fh._get_channel(key=key)
        reader.get_counts.assert_not_called()

        reader.get_calibrated_channels.reset_mock()
        for calib in ['reflectance', 'brightness_temperature']:
            key = make_dataid(name='1', calibration=calib)
            fh._get_channel(key)
            reader.get_calibrated_channels.assert_not_called()

    def test_get_angle(self):
        """Test getting the angle."""
        from satpy.tests.utils import make_dataid

        reader = mock.MagicMock()
        reader.get_angles.return_value = 1, 2, 3, 4, 5
        fh = _get_fh_mocked(reader=reader, angles=None)

        # Test angle readout
        key = make_dataid(name='sensor_zenith_angle')
        res = fh._get_angle(key)
        self.assertEqual(res, 2)
        self.assertDictEqual(fh.angles, {'sensor_zenith_angle': 2,
                                         'sensor_azimuth_angle': 1,
                                         'solar_zenith_angle': 4,
                                         'solar_azimuth_angle': 3,
                                         'sun_sensor_azimuth_difference_angle': 5})

        # Test buffering
        key = make_dataid(name='sensor_azimuth_angle')
        fh._get_angle(key)
        reader.get_angles.assert_called_once()

    def test_strip_invalid_lat(self):
        """Test stripping invalid coordinates."""
        import pygac.utils

        reader = mock.MagicMock()
        reader.get_lonlat.return_value = None, None
        fh = _get_fh_mocked(reader=reader, first_valid_lat=None)

        # Test stripping
        pygac.utils.strip_invalid_lat.return_value = 1, 2
        start, end = fh._strip_invalid_lat()
        self.assertTupleEqual((start, end), (1, 2))

        # Test buffering
        fh._strip_invalid_lat()
        pygac.utils.strip_invalid_lat.assert_called_once()

    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile._slice')
    def test_slice(self, _slice):
        """Test slicing."""

        def _slice_patched(data):
            return data[1:3]
        _slice.side_effect = _slice_patched

        data = np.zeros((4, 2))
        times = np.array([1, 2, 3, 4], dtype='datetime64[us]')

        fh = _get_fh_mocked(start_line=1, end_line=3, strip_invalid_coords=False)
        data_slc, times_slc = fh.slice(data, times)
        np.testing.assert_array_equal(data_slc, data[1:3])
        np.testing.assert_array_equal(times_slc, times[1:3])
        self.assertEqual(fh.start_time, datetime(1970, 1, 1, 0, 0, 0, 2))
        self.assertEqual(fh.end_time, datetime(1970, 1, 1, 0, 0, 0, 3))

    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile._get_qual_flags')
    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile._strip_invalid_lat')
    def test__slice(self, strip_invalid_lat, get_qual_flags):
        """Test slicing."""
        import pygac.utils
        pygac.utils.check_user_scanlines.return_value = 1, 2
        pygac.utils.slice_channel.return_value = 'sliced'
        strip_invalid_lat.return_value = 3, 4
        get_qual_flags.return_value = 'qual_flags'

        data = np.zeros((2, 2))

        # a) Only start/end line given
        fh = _get_fh_mocked(start_line=5, end_line=6, strip_invalid_coords=False)
        data_slc = fh._slice(data)
        self.assertEqual(data_slc, 'sliced')
        pygac.utils.check_user_scanlines.assert_called_with(
            start_line=5, end_line=6,
            first_valid_lat=None, last_valid_lat=None, along_track=2)
        pygac.utils.slice_channel.assert_called_with(
            data, start_line=1, end_line=2,
            first_valid_lat=None, last_valid_lat=None)

        # b) Only strip_invalid_coords=True
        fh = _get_fh_mocked(start_line=None, end_line=None, strip_invalid_coords=True)
        fh._slice(data)
        pygac.utils.check_user_scanlines.assert_called_with(
            start_line=0, end_line=0,
            first_valid_lat=3, last_valid_lat=4, along_track=2)

        # c) Both
        fh = _get_fh_mocked(start_line=5, end_line=6, strip_invalid_coords=True)
        fh._slice(data)
        pygac.utils.check_user_scanlines.assert_called_with(
            start_line=5, end_line=6,
            first_valid_lat=3, last_valid_lat=4, along_track=2)

        # Test slicing with older pygac versions
        pygac.utils.slice_channel.return_value = ('sliced', 'foo', 'bar')
        data_slc = fh._slice(data)
        self.assertEqual(data_slc, 'sliced')


class TestGetDataset(GACLACFilePatcher):
    """Test the get_dataset method."""

    def setUp(self):
        """Set up the instance."""
        self.exp = xr.DataArray(da.ones((3, 3)),
                                name='1',
                                dims=('y', 'x'),
                                coords={'acq_time': ('y', [0, 1, 2])},
                                attrs={'name': '1',
                                       'platform_name': 'spacecraft_name',
                                       'orbit_number': 123,
                                       'sensor': 'sensor',
                                       'foo': 'bar',
                                       'standard_name': 'my_standard_name'})
        self.exp.coords['acq_time'].attrs['long_name'] = 'Mean scanline acquisition time'
        super().setUp()

    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile.__init__', return_value=None)
    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile.read_raw_data')
    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile._get_channel', return_value=np.ones((3, 3)))
    def test_get_dataset_channels(self, get_channel, *mocks):
        """Test getting the channel datasets."""
        pygac_reader = _get_reader_mocked()
        fh = self._create_file_handler(pygac_reader)

        # Test calibration to reflectance as well as attributes.
        key, res = self._get_dataset(fh)
        exp = self._create_expected(res.name)
        exp.attrs['orbital_parameters'] = {'tle': 'tle'}

        xr.testing.assert_identical(res, exp)
        get_channel.assert_called_with(key)

        self._check_get_channel_calls(fh, get_channel)

    @staticmethod
    def _get_dataset(fh):
        from satpy.tests.utils import make_dataid

        key = make_dataid(name='1', calibration='reflectance')
        info = {'name': '1', 'standard_name': 'my_standard_name'}
        res = fh.get_dataset(key=key, info=info)
        return key, res

    @staticmethod
    def _create_file_handler(reader):
        """Mock reader and file handler."""
        fh = _get_fh_mocked(
            reader=reader,
            chn_dict={'1': 0, '5': 0},
            start_line=None,
            end_line=None,
            strip_invalid_coords=False,
            filename_info={'orbit_number': 123},
            sensor='sensor',
        )
        return fh

    @staticmethod
    def _create_expected(name):
        exp = xr.DataArray(da.ones((3, 3)),
                           name=name,
                           dims=('y', 'x'),
                           coords={'acq_time': ('y', [0, 1, 2])},
                           attrs={'name': '1',
                                  'platform_name': 'spacecraft_name',
                                  'orbit_number': 123,
                                  'sensor': 'sensor',
                                  'foo': 'bar',
                                  'standard_name': 'my_standard_name'})
        exp.coords['acq_time'].attrs['long_name'] = 'Mean scanline acquisition time'
        return exp

    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile.__init__', return_value=None)
    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile.read_raw_data')
    @mock.patch('satpy.readers.avhrr_l1b_gaclac.GACLACFile._get_channel', return_value=np.ones((3, 3)))
    def test_get_dataset_no_tle(self, get_channel, *mocks):
        """Test getting the channel datasets when no TLEs are present."""
        pygac_reader = _get_reader_mocked()
        pygac_reader.get_tle_lines = mock.MagicMock()
        pygac_reader.get_tle_lines.side_effect = RuntimeError()

        fh = self._create_file_handler(pygac_reader)

        # Test calibration to reflectance as well as attributes.
        key, res = self._get_dataset(fh)
        exp = self._create_expected(res.name)
        xr.testing.assert_identical(res, exp)
        get_channel.assert_called_with(key)

        self._check_get_channel_calls(fh, get_channel)

    @staticmethod
    def _check_get_channel_calls(fh, get_channel):
        """Check _get_channel() calls."""
        from satpy.tests.utils import make_dataid

        for key in [make_dataid(name='1', calibration='counts'),
                    make_dataid(name='5', calibration='brightness_temperature')]:
            fh.get_dataset(key=key, info={'name': 1})
            get_channel.assert_called_with(key)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2019-2020 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.

"""Reader for the SEVIRI L2 products in GRIB2 format.

References:
    FM 92 GRIB Edition 2
    https://www.wmo.int/pages/prog/www/WMOCodes/Guides/GRIB/GRIB2_062006.pdf
    EUMETSAT Product Navigator
    https://navigator.eumetsat.int/
"""

import logging
from datetime import timedelta

import dask.array as da
import numpy as np
import xarray as xr

from satpy import CHUNK_SIZE
from satpy.readers._geos_area import get_area_definition, get_geos_area_naming
from satpy.readers.eum_base import get_service_mode
from satpy.readers.file_handlers import BaseFileHandler
from satpy.readers.seviri_base import PLATFORM_DICT, REPEAT_CYCLE_DURATION, calculate_area_extent

try:
    import eccodes as ec
except ImportError:
    raise ImportError(
        "Missing eccodes-python and/or eccodes C-library installation. Use conda to install eccodes")

logger = logging.getLogger(__name__)


class SeviriL2GribFileHandler(BaseFileHandler):
    """Reader class for SEVIRI L2 products in GRIB format."""

    def __init__(self, filename, filename_info, filetype_info):
        """Read the global attributes and prepare for dataset reading."""
        super().__init__(filename, filename_info, filetype_info)
        # Turn on support for multiple fields in single GRIB messages (required for SEVIRI L2 files)
        ec.codes_grib_multi_support_on()

    @property
    def start_time(self):
        """Return the sensing start time."""
        return self.filename_info['start_time']

    @property
    def end_time(self):
        """Return the sensing end time."""
        return self.start_time + timedelta(minutes=REPEAT_CYCLE_DURATION)

    def get_area_def(self, dataset_id):
        """Return the area definition for a dataset."""
        self._area_dict['column_step'] = dataset_id.resolution
        self._area_dict['line_step'] = dataset_id.resolution

        area_extent = calculate_area_extent(self._area_dict)

        # Call the get_area_definition function to obtain the area
        area_def = get_area_definition(self._pdict, area_extent)

        return area_def

    def get_dataset(self, dataset_id, dataset_info):
        """Get dataset using the parameter_number key in dataset_info.

        In a previous version of the reader, the attributes (nrows, ncols, ssp_lon) and projection information
        (pdict and area_dict) were computed while initializing the file handler. Also the code would break out from
        the While-loop below as soon as the correct parameter_number was found. This has now been revised becasue the
        reader would sometimes give corrupt information about the number of messages in the file and the dataset
        dimensions within a given message if the file was only partly read (not looping over all messages) in an earlier
        instance.
        """
        logger.debug('Reading in file to get dataset with parameter number %d.',
                     dataset_info['parameter_number'])

        xarr = None
        message_found = False
        with open(self.filename, 'rb') as fh:

            # Iterate over all messages and fetch data when the correct parameter number is found
            while True:
                gid = ec.codes_grib_new_from_file(fh)

                if gid is None:
                    if not message_found:
                        # Could not obtain a valid message ID from the grib file
                        logger.warning("Could not find parameter_number %d in GRIB file, no valid Dataset created",
                                       dataset_info['parameter_number'])
                    break

                # Check if the parameter number in the GRIB message corresponds to the required key
                parameter_number = self._get_from_msg(gid, 'parameterNumber')

                if parameter_number == dataset_info['parameter_number']:

                    self._res = dataset_id.resolution
                    self._read_attributes(gid)

                    # Read the missing value
                    missing_value = self._get_from_msg(gid, 'missingValue')

                    # Retrieve values and metadata from the GRIB message, masking the values equal to missing_value
                    xarr = self._get_xarray_from_msg(gid)

                    xarr.data = da.where(xarr.data == missing_value, np.nan, xarr.data)

                    ec.codes_release(gid)

                    # Combine all metadata into the dataset attributes and break out of the loop
                    xarr.attrs.update(dataset_info)
                    xarr.attrs.update(self._get_attributes())

                    message_found = True

                else:
                    # The parameter number is not the correct one, release gid and skip to next message
                    ec.codes_release(gid)

        return xarr

    def _read_attributes(self, gid):
        """Read the parameter attributes from the message and create the projection and area dictionaries."""
        # Read SSP and date/time
        self._ssp_lon = self._get_from_msg(gid, 'longitudeOfSubSatellitePointInDegrees')

        # Read number of points on the x and y axes
        self._nrows = self._get_from_msg(gid, 'Ny')
        self._ncols = self._get_from_msg(gid, 'Nx')

        # Creates the projection and area dictionaries
        self._pdict, self._area_dict = self._get_proj_area(gid)

    def _get_proj_area(self, gid):
        """Compute the dictionary with the projection and area definition from a GRIB message.

        Args:
            gid: The ID of the GRIB message.
        Returns:
            tuple: A tuple of two dictionaries for the projection and the area definition.
                pdict:
                    a: Earth major axis [m]
                    b: Earth minor axis [m]
                    h: Height over surface [m]
                    ssp_lon: longitude of subsatellite point [deg]
                    nlines: number of lines
                    ncols: number of columns
                    a_name: name of the area
                    a_desc: description of the area
                    p_id: id of the projection
                area_dict:
                    center_point: coordinate of the center point
                    north: coodinate of the north limit
                    east: coodinate of the east limit
                    west: coodinate of the west limit
                    south: coodinate of the south limit
        """
        # Get name of area definition
        area_naming_input_dict = {'platform_name': 'msg',
                                  'instrument_name': 'seviri',
                                  'resolution': self._res,
                                  }

        area_naming = get_geos_area_naming({**area_naming_input_dict,
                                            **get_service_mode('seviri', self._ssp_lon)})

        # Read all projection and area parameters from the message
        earth_major_axis_in_meters = self._get_from_msg(gid, 'earthMajorAxis') * 1000.0  # [m]
        earth_minor_axis_in_meters = self._get_from_msg(gid, 'earthMinorAxis') * 1000.0  # [m]

        earth_major_axis_in_meters = self._scale_earth_axis(earth_major_axis_in_meters)
        earth_minor_axis_in_meters = self._scale_earth_axis(earth_minor_axis_in_meters)

        nr_in_radius_of_earth = self._get_from_msg(gid, 'NrInRadiusOfEarth')
        xp_in_grid_lengths = self._get_from_msg(gid, 'XpInGridLengths')
        h_in_meters = earth_major_axis_in_meters * (nr_in_radius_of_earth - 1.0)  # [m]

        # Create the dictionary with the projection data
        pdict = {
            'a': earth_major_axis_in_meters,
            'b': earth_minor_axis_in_meters,
            'h': h_in_meters,
            'ssp_lon': self._ssp_lon,
            'nlines': self._ncols,
            'ncols': self._nrows,
            'a_name': area_naming['area_id'],
            'a_desc': area_naming['description'],
            'p_id': "",
        }

        # Compute the dictionary with the area extension
        area_dict = {
            'center_point': xp_in_grid_lengths,
            'north': self._nrows,
            'east': 1,
            'west': self._ncols,
            'south': 1,
        }

        return pdict, area_dict

    @staticmethod
    def _scale_earth_axis(data):
        """Scale Earth axis data to make sure the value matched the expected unit [m].

        The earthMinorAxis value stored in the aerosol over sea product is scaled incorrectly by a factor of 1e8. This
        method provides a flexible temporarily workaraound by making sure that all earth axis values are scaled such
        that they are on the order of millions of meters as expected by the reader. As soon as the scaling issue has
        been resolved by EUMETSAT this workaround can be removed.

        """
        scale_factor = 10 ** np.ceil(np.log10(1e6/data))
        return data * scale_factor

    def _get_xarray_from_msg(self, gid):
        """Read the values from the GRIB message and return a DataArray object.

        Args:
            gid: The ID of the GRIB message.
        Returns:
            DataArray: The array containing the retrieved values.
        """
        # Data from GRIB message are read into an Xarray...
        xarr = xr.DataArray(da.from_array(ec.codes_get_values(
            gid).reshape(self._nrows, self._ncols), CHUNK_SIZE), dims=('y', 'x'))

        return xarr

    def _get_attributes(self):
        """Create a dictionary of  attributes to be added to the dataset.

        Returns:
            dict: A dictionary of parameter attributes.
                ssp_lon: longitude of subsatellite point
                sensor: name of sensor
                platform_name: name of the platform
        """
        orbital_parameters = {
            'projection_longitude': self._ssp_lon
        }

        attributes = {
            'orbital_parameters': orbital_parameters,
            'sensor': 'seviri',
            'platform_name': PLATFORM_DICT[self.filename_info['spacecraft']]
        }
        return attributes

    @staticmethod
    def _get_from_msg(gid, key):
        """Get a value from the GRIB message based on the key, return None if missing.

        Args:
            gid: The ID of the GRIB message.
            key: The key of the required attribute.
        Returns:
            The retrieved attribute or None if the key is missing.
        """
        try:
            attr = ec.codes_get(gid, key)
        except ec.KeyValueNotFoundError:
            logger.warning("Key %s not found in GRIB message", key)
            attr = None
        return attr



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2017 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Unit testing the enhancements functions, e.g. cira_stretch."""

import contextlib
import os
from tempfile import NamedTemporaryFile
from unittest import mock

import dask.array as da
import numpy as np
import pytest
import xarray as xr

from satpy.enhancements import create_colormap, on_dask_array, on_separate_bands, using_map_blocks


def run_and_check_enhancement(func, data, expected, **kwargs):
    """Perform basic checks that apply to multiple tests."""
    from trollimage.xrimage import XRImage

    pre_attrs = data.attrs
    img = XRImage(data)
    func(img, **kwargs)

    assert isinstance(img.data.data, da.Array)
    old_keys = set(pre_attrs.keys())
    # It is OK to have "enhancement_history" added
    new_keys = set(img.data.attrs.keys()) - {"enhancement_history"}
    assert old_keys == new_keys

    res_data_arr = img.data
    assert isinstance(res_data_arr, xr.DataArray)
    assert isinstance(res_data_arr.data, da.Array)
    res_data = res_data_arr.data.compute()  # mimics what xrimage geotiff writing does
    assert not isinstance(res_data, da.Array)
    np.testing.assert_allclose(res_data, expected, atol=1.e-6, rtol=0)


def identical_decorator(func):
    """Decorate but do nothing."""
    return func


class TestEnhancementStretch:
    """Class for testing enhancements in satpy.enhancements."""

    def setup_method(self):
        """Create test data used by every test."""
        data = np.arange(-210, 790, 100).reshape((2, 5)) * 0.95
        data[0, 0] = np.nan  # one bad value for testing
        crefl_data = np.arange(-210, 790, 100).reshape((2, 5)) * 0.95
        crefl_data /= 5.605
        crefl_data[0, 0] = np.nan  # one bad value for testing
        crefl_data[0, 1] = 0.
        self.ch1 = xr.DataArray(da.from_array(data, chunks=2), dims=('y', 'x'), attrs={'test': 'test'})
        self.ch2 = xr.DataArray(da.from_array(crefl_data, chunks=2), dims=('y', 'x'), attrs={'test': 'test'})
        rgb_data = np.stack([data, data, data])
        self.rgb = xr.DataArray(da.from_array(rgb_data, chunks=(3, 2, 2)),
                                dims=('bands', 'y', 'x'),
                                coords={'bands': ['R', 'G', 'B']})

    @pytest.mark.parametrize(
        ("decorator", "exp_call_cls"),
        [
            (identical_decorator, xr.DataArray),
            (on_dask_array, da.Array),
            (using_map_blocks, np.ndarray),
        ],
    )
    @pytest.mark.parametrize("input_data_name", ["ch1", "ch2", "rgb"])
    def test_apply_enhancement(self, input_data_name, decorator, exp_call_cls):
        """Test the 'apply_enhancement' utility function."""
        def _enh_func(img):
            def _calc_func(data):
                assert isinstance(data, exp_call_cls)
                return data
            decorated_func = decorator(_calc_func)
            return decorated_func(img.data)

        in_data = getattr(self, input_data_name)
        exp_data = in_data.values
        if "bands" not in in_data.coords:
            exp_data = exp_data[np.newaxis, :, :]
        run_and_check_enhancement(_enh_func, in_data, exp_data)

    def test_cira_stretch(self):
        """Test applying the cira_stretch."""
        from satpy.enhancements import cira_stretch

        expected = np.array([[
            [np.nan, -7.04045974, -7.04045974, 0.79630132, 0.95947296],
            [1.05181359, 1.11651012, 1.16635571, 1.20691137, 1.24110186]]])
        run_and_check_enhancement(cira_stretch, self.ch1, expected)

    def test_reinhard(self):
        """Test the reinhard algorithm."""
        from satpy.enhancements import reinhard_to_srgb
        expected = np.array([[[np.nan, 0., 0., 0.93333793, 1.29432402],
                              [1.55428709, 1.76572249, 1.94738635, 2.10848544, 2.25432809]],

                             [[np.nan, 0., 0., 0.93333793, 1.29432402],
                              [1.55428709, 1.76572249, 1.94738635, 2.10848544, 2.25432809]],

                             [[np.nan, 0., 0., 0.93333793, 1.29432402],
                              [1.55428709, 1.76572249, 1.94738635, 2.10848544, 2.25432809]]])
        run_and_check_enhancement(reinhard_to_srgb, self.rgb, expected)

    def test_lookup(self):
        """Test the lookup enhancement function."""
        from satpy.enhancements import lookup
        expected = np.array([[
            [0., 0., 0., 0.333333, 0.705882],
            [1., 1., 1., 1., 1.]]])
        lut = np.arange(256.)
        run_and_check_enhancement(lookup, self.ch1, expected, luts=lut)

        expected = np.array([[[0., 0., 0., 0.333333, 0.705882],
                              [1., 1., 1., 1., 1.]],
                             [[0., 0., 0., 0.333333, 0.705882],
                              [1., 1., 1., 1., 1.]],
                             [[0., 0., 0., 0.333333, 0.705882],
                              [1., 1., 1., 1., 1.]]])
        lut = np.arange(256.)
        lut = np.vstack((lut, lut, lut)).T
        run_and_check_enhancement(lookup, self.rgb, expected, luts=lut)

    def test_colorize(self):
        """Test the colorize enhancement function."""
        from trollimage.colormap import brbg

        from satpy.enhancements import colorize
        expected = np.array([[
            [np.nan, 3.29409498e-01, 3.29409498e-01,
             4.35952940e-06, 4.35952940e-06],
            [4.35952940e-06, 4.35952940e-06, 4.35952940e-06,
             4.35952940e-06, 4.35952940e-06]],
            [[np.nan, 1.88249866e-01, 1.88249866e-01,
              2.35302110e-01, 2.35302110e-01],
             [2.35302110e-01, 2.35302110e-01, 2.35302110e-01,
              2.35302110e-01, 2.35302110e-01]],
            [[np.nan, 1.96102817e-02, 1.96102817e-02,
              1.88238767e-01, 1.88238767e-01],
             [1.88238767e-01, 1.88238767e-01, 1.88238767e-01,
              1.88238767e-01, 1.88238767e-01]]])
        run_and_check_enhancement(colorize, self.ch1, expected, palettes=brbg)

    def test_palettize(self):
        """Test the palettize enhancement function."""
        from trollimage.colormap import brbg

        from satpy.enhancements import palettize
        expected = np.array([[[10, 0, 0, 10, 10], [10, 10, 10, 10, 10]]])
        run_and_check_enhancement(palettize, self.ch1, expected, palettes=brbg)

    def test_three_d_effect(self):
        """Test the three_d_effect enhancement function."""
        from satpy.enhancements import three_d_effect
        expected = np.array([[
            [np.nan, np.nan, -389.5, -294.5, 826.5],
            [np.nan, np.nan, 85.5, 180.5, 1301.5]]])
        run_and_check_enhancement(three_d_effect, self.ch1, expected)

    def test_crefl_scaling(self):
        """Test the crefl_scaling enhancement function."""
        from satpy.enhancements import crefl_scaling
        expected = np.array([[
            [np.nan, 0., 0., 0.44378, 0.631734],
            [0.737562, 0.825041, 0.912521, 1., 1.]]])
        run_and_check_enhancement(crefl_scaling, self.ch2, expected, idx=[0., 25., 55., 100., 255.],
                                  sc=[0., 90., 140., 175., 255.])

    def test_piecewise_linear_stretch(self):
        """Test the piecewise_linear_stretch enhancement function."""
        from satpy.enhancements import piecewise_linear_stretch
        expected = np.array([[
            [np.nan, 0., 0., 0.44378, 0.631734],
            [0.737562, 0.825041, 0.912521, 1., 1.]]])
        run_and_check_enhancement(piecewise_linear_stretch,
                                  self.ch2 / 100.0,
                                  expected,
                                  xp=[0., 25., 55., 100., 255.],
                                  fp=[0., 90., 140., 175., 255.],
                                  reference_scale_factor=255,
                                  )

    def test_btemp_threshold(self):
        """Test applying the cira_stretch."""
        from satpy.enhancements import btemp_threshold

        expected = np.array([[
            [np.nan, 0.946207, 0.892695, 0.839184, 0.785672],
            [0.73216, 0.595869, 0.158745, -0.278379, -0.715503]]])
        run_and_check_enhancement(btemp_threshold, self.ch1, expected,
                                  min_in=-200, max_in=500, threshold=350)

    def test_merge_colormaps(self):
        """Test merging colormaps."""
        from trollimage.colormap import Colormap

        from satpy.enhancements import _merge_colormaps as mcp
        from satpy.enhancements import create_colormap
        ret_map = mock.MagicMock()

        create_colormap_mock = mock.Mock(wraps=create_colormap)
        cmap1 = Colormap((1, (1., 1., 1.)))
        kwargs = {'palettes': cmap1}

        with mock.patch('satpy.enhancements.create_colormap', create_colormap_mock):
            res = mcp(kwargs)
        assert res is cmap1
        create_colormap_mock.assert_not_called()
        create_colormap_mock.reset_mock()
        ret_map.reset_mock()

        cmap1 = {'colors': 'blues', 'min_value': 0,
                 'max_value': 1}
        kwargs = {'palettes': [cmap1]}
        with mock.patch('satpy.enhancements.create_colormap', create_colormap_mock),\
                mock.patch('trollimage.colormap.blues', ret_map):
            _ = mcp(kwargs)
        create_colormap_mock.assert_called_once()
        ret_map.reverse.assert_not_called()
        ret_map.set_range.assert_called_with(0, 1)
        create_colormap_mock.reset_mock()
        ret_map.reset_mock()

        cmap2 = {'colors': 'blues', 'min_value': 2,
                 'max_value': 3, 'reverse': True}
        kwargs = {'palettes': [cmap2]}
        with mock.patch('trollimage.colormap.blues', ret_map):
            _ = mcp(kwargs)
        ret_map.reverse.assert_called_once()
        ret_map.set_range.assert_called_with(2, 3)
        create_colormap_mock.reset_mock()
        ret_map.reset_mock()

        kwargs = {'palettes': [cmap1, cmap2]}
        with mock.patch('trollimage.colormap.blues', ret_map):
            _ = mcp(kwargs)
        ret_map.__add__.assert_called_once()

    def tearDown(self):
        """Clean up."""


@contextlib.contextmanager
def closed_named_temp_file(**kwargs):
    """Named temporary file context manager that closes the file after creation.

    This helps with Windows systems which can get upset with opening or
    deleting a file that is already open.

    """
    try:
        with NamedTemporaryFile(delete=False, **kwargs) as tmp_cmap:
            yield tmp_cmap.name
    finally:
        os.remove(tmp_cmap.name)


def _write_cmap_to_file(cmap_filename, cmap_data):
    ext = os.path.splitext(cmap_filename)[1]
    if ext in (".npy",):
        np.save(cmap_filename, cmap_data)
    elif ext in (".npz",):
        np.savez(cmap_filename, cmap_data)
    else:
        np.savetxt(cmap_filename, cmap_data, delimiter=",")


def _generate_cmap_test_data(color_scale, colormap_mode):
    cmap_data = np.array([
        [1, 0, 0],
        [1, 1, 0],
        [1, 1, 1],
        [0, 0, 1],
    ], dtype=np.float64)
    if len(colormap_mode) != 3:
        _cmap_data = cmap_data
        cmap_data = np.empty((cmap_data.shape[0], len(colormap_mode)),
                             dtype=np.float64)
        if colormap_mode.startswith("V") or colormap_mode.endswith("A"):
            cmap_data[:, 0] = np.array([128, 130, 132, 134]) / 255.0
            cmap_data[:, -3:] = _cmap_data
        if colormap_mode.startswith("V") and colormap_mode.endswith("A"):
            cmap_data[:, 1] = np.array([128, 130, 132, 134]) / 255.0
    if color_scale is None or color_scale == 255:
        cmap_data = (cmap_data * 255).astype(np.uint8)
    return cmap_data


class TestColormapLoading:
    """Test utilities used with colormaps."""

    @pytest.mark.parametrize("color_scale", [None, 1.0])
    @pytest.mark.parametrize("colormap_mode", ["RGB", "VRGB", "VRGBA"])
    @pytest.mark.parametrize("extra_kwargs",
                             [
                                 {},
                                 {"min_value": 50, "max_value": 100},
                             ])
    @pytest.mark.parametrize("filename_suffix", [".npy", ".npz", ".csv"])
    def test_cmap_from_file(self, color_scale, colormap_mode, extra_kwargs, filename_suffix):
        """Test that colormaps can be loaded from a binary file."""
        # create the colormap file on disk
        with closed_named_temp_file(suffix=filename_suffix) as cmap_filename:
            cmap_data = _generate_cmap_test_data(color_scale, colormap_mode)
            _write_cmap_to_file(cmap_filename, cmap_data)

            unset_first_value = 128.0 / 255.0 if colormap_mode.startswith("V") else 0.0
            unset_last_value = 134.0 / 255.0 if colormap_mode.startswith("V") else 1.0
            if (color_scale is None or color_scale == 255) and colormap_mode.startswith("V"):
                unset_first_value *= 255
                unset_last_value *= 255
            if "min_value" in extra_kwargs:
                unset_first_value = extra_kwargs["min_value"]
                unset_last_value = extra_kwargs["max_value"]

            first_color = [1.0, 0.0, 0.0]
            if colormap_mode == "VRGBA":
                first_color = [128.0 / 255.0] + first_color

            kwargs1 = {"filename": cmap_filename}
            kwargs1.update(extra_kwargs)
            if color_scale is not None:
                kwargs1["color_scale"] = color_scale

            cmap = create_colormap(kwargs1)
            assert cmap.colors.shape[0] == 4
            np.testing.assert_equal(cmap.colors[0], first_color)
            assert cmap.values.shape[0] == 4
            assert cmap.values[0] == unset_first_value
            assert cmap.values[-1] == unset_last_value

    def test_cmap_vrgb_as_rgba(self):
        """Test that data created as VRGB still reads as RGBA."""
        with closed_named_temp_file(suffix=".npy") as cmap_filename:
            cmap_data = _generate_cmap_test_data(None, "VRGB")
            np.save(cmap_filename, cmap_data)
            cmap = create_colormap({'filename': cmap_filename, 'colormap_mode': "RGBA"})
            assert cmap.colors.shape[0] == 4
            assert cmap.colors.shape[1] == 4  # RGBA
            np.testing.assert_equal(cmap.colors[0], [128 / 255., 1.0, 0, 0])
            assert cmap.values.shape[0] == 4
            assert cmap.values[0] == 0
            assert cmap.values[-1] == 1.0

    @pytest.mark.parametrize(
        ("real_mode", "forced_mode"),
        [
            ("VRGBA", "RGBA"),
            ("VRGBA", "VRGB"),
            ("RGBA", "RGB"),
        ]
    )
    @pytest.mark.parametrize("filename_suffix", [".npy", ".csv"])
    def test_cmap_bad_mode(self, real_mode, forced_mode, filename_suffix):
        """Test that reading colormaps with the wrong mode fails."""
        with closed_named_temp_file(suffix=filename_suffix) as cmap_filename:
            cmap_data = _generate_cmap_test_data(None, real_mode)
            _write_cmap_to_file(cmap_filename, cmap_data)
            # Force colormap_mode VRGBA to RGBA and we should see an exception
            with pytest.raises(ValueError):
                create_colormap({'filename': cmap_filename, 'colormap_mode': forced_mode})

    def test_cmap_from_file_bad_shape(self):
        """Test that unknown array shape causes an error."""
        from satpy.enhancements import create_colormap

        # create the colormap file on disk
        with closed_named_temp_file(suffix='.npy') as cmap_filename:
            np.save(cmap_filename, np.array([
                [0],
                [64],
                [128],
                [255],
            ]))

            with pytest.raises(ValueError):
                create_colormap({'filename': cmap_filename})

    def test_cmap_from_config_path(self, tmp_path):
        """Test loading a colormap relative to a config path."""
        import satpy
        from satpy.enhancements import create_colormap

        cmap_dir = tmp_path / "colormaps"
        cmap_dir.mkdir()
        cmap_filename = cmap_dir / "my_colormap.npy"
        cmap_data = _generate_cmap_test_data(None, "RGBA")
        np.save(cmap_filename, cmap_data)
        with satpy.config.set(config_path=[tmp_path]):
            rel_cmap_filename = os.path.join("colormaps", "my_colormap.npy")
            cmap = create_colormap({'filename': rel_cmap_filename, 'colormap_mode': "RGBA"})
            assert cmap.colors.shape[0] == 4
            assert cmap.colors.shape[1] == 4  # RGBA
            np.testing.assert_equal(cmap.colors[0], [128 / 255., 1.0, 0, 0])
            assert cmap.values.shape[0] == 4
            assert cmap.values[0] == 0
            assert cmap.values[-1] == 1.0

    def test_cmap_from_trollimage(self):
        """Test that colormaps in trollimage can be loaded."""
        from satpy.enhancements import create_colormap
        cmap = create_colormap({'colors': 'pubu'})
        from trollimage.colormap import pubu
        np.testing.assert_equal(cmap.colors, pubu.colors)
        np.testing.assert_equal(cmap.values, pubu.values)

    def test_cmap_no_colormap(self):
        """Test that being unable to create a colormap raises an error."""
        from satpy.enhancements import create_colormap
        with pytest.raises(ValueError):
            create_colormap({})

    def test_cmap_list(self):
        """Test that colors can be a list/tuple."""
        from satpy.enhancements import create_colormap
        colors = [
            [0, 0, 1],
            [1, 0, 1],
            [0, 1, 1],
            [1, 1, 1],
        ]
        values = [2, 4, 6, 8]
        cmap = create_colormap({'colors': colors, 'color_scale': 1})
        assert cmap.colors.shape[0] == 4
        np.testing.assert_equal(cmap.colors[0], [0.0, 0.0, 1.0])
        assert cmap.values.shape[0] == 4
        assert cmap.values[0] == 0
        assert cmap.values[-1] == 1.0

        cmap = create_colormap({'colors': colors, 'color_scale': 1, 'values': values})
        assert cmap.colors.shape[0] == 4
        np.testing.assert_equal(cmap.colors[0], [0.0, 0.0, 1.0])
        assert cmap.values.shape[0] == 4
        assert cmap.values[0] == 2
        assert cmap.values[-1] == 8


def test_on_separate_bands():
    """Test the `on_separate_bands` decorator."""
    def func(array, index, gain=2):
        return xr.DataArray(np.ones(array.shape, dtype=array.dtype) * index * gain,
                            coords=array.coords, dims=array.dims, attrs=array.attrs)

    separate_func = on_separate_bands(func)
    arr = xr.DataArray(np.zeros((3, 10, 10)), dims=['bands', 'y', 'x'], coords={"bands": ["R", "G", "B"]})
    assert separate_func(arr).shape == arr.shape
    assert all(separate_func(arr, gain=1).values[:, 0, 0] == [0, 1, 2])


def test_using_map_blocks():
    """Test the `using_map_blocks` decorator."""
    def func(np_array, block_info=None):
        value = block_info[0]['chunk-location'][-1]
        return np.ones(np_array.shape) * value

    map_blocked_func = using_map_blocks(func)
    arr = xr.DataArray(da.zeros((3, 10, 10), dtype=int, chunks=5), dims=['bands', 'y', 'x'])
    res = map_blocked_func(arr)
    assert res.shape == arr.shape
    assert res[0, 0, 0].compute() != res[0, 9, 9].compute()


def test_on_dask_array():
    """Test the `on_dask_array` decorator."""
    def func(dask_array):
        if not isinstance(dask_array, da.core.Array):
            pytest.fail("Array is not a dask array")
        return dask_array

    dask_func = on_dask_array(func)
    arr = xr.DataArray(da.zeros((3, 10, 10), dtype=int, chunks=5), dims=['bands', 'y', 'x'])
    res = dask_func(arr)
    assert res.shape == arr.shape



# Copyright (c) 2022 Satpy developers
#
# satpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# satpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Common utility modules used for LI mock-oriented unit tests."""

from datetime import datetime

import numpy as np
import xarray as xr

from satpy.tests.reader_tests.test_netcdf_utils import FakeNetCDF4FileHandler

# mapping of netcdf type code to numpy data type:
TYPE_MAP = {
    'i1': np.int8,
    'i2': np.int16,
    'i4': np.int32,
    'i8': np.int64,
    'u1': np.uint8,
    'u2': np.uint16,
    'u4': np.uint32,
    'u8': np.uint64,
    'f4': np.float32,
    'f8': np.float64,
}


def l2_le_schema(settings=None):
    """Define schema for LI L2 LE product."""
    settings = settings or {}
    nobs = settings.get("num_obs", 123)
    nchunks = settings.get("num_chunks", 23)
    nfilters = settings.get("num_filters", 2)

    def rand_u16(num):
        return np.random.randint(low=0, high=np.iinfo(np.uint16).max - 1, size=num, dtype=np.uint16)

    return {
        'providers': settings.get('providers', {}),
        'variable_path': settings.get('variable_path', 'data/'),
        'dimensions': {
            'unfiltered_events': nobs,
            'l1b_chunks': nchunks,
            'l1b_offsets': nchunks,
            'filters': nfilters,
            'scalar': 1,
        },
        'variables': {},
        'sector_variables': {
            "event_id": {
                "format": "u2",
                "shape": ('unfiltered_events',),
                "fill_value": 65535,
                "long_name": "ID of LI L2 Event",
                "default_data": lambda: rand_u16(nobs)
            },
            "group_id": {
                "format": "u2",
                "shape": ('unfiltered_events',),
                "fill_value": 65535,
                "long_name": "ID of associated LI L2 Group object",
                "default_data": lambda: rand_u16(nobs)
            },
            "l1b_chunk_ids": {
                "format": "u4",
                "shape": ('l1b_chunks',),
                "fill_value": 4294967295,
                "long_name": "Array of L1b event chunk IDs",
                "default_data": lambda: np.arange(nchunks) + 10000
            },
            "l1b_chunk_offsets": {
                "format": "u4",
                "shape": ('l1b_offsets',),
                "fill_value": 4294967295,
                "long_name": "Array offset for L1b event chunk boundaries",
                "default_data": lambda: np.arange(nchunks)
            },
            "l1b_window": {
                "format": "u4",
                "shape": ('unfiltered_events',),
                "fill_value": 4294967295,
                "long_name": "window index of associated L1b event",
                "default_data": lambda: (np.arange(nobs) + 10000)
            },
            "filter_values": {
                "format": "u1",
                "shape": ('unfiltered_events', 'filters',),
                "fill_value": 255,
                "scale_factor": 0.004,
                "add_offset": 0.0,
                "long_name": "L2 filter results",
                "default_data": lambda: np.random.randint(low=0, high=255, size=(nobs, nfilters), dtype=np.uint8)
            },
            "epoch_time": {
                "format": "f8",
                "shape": ('scalar',),
                "fill_value": 9.96920996886869e36,
                "long_name": "Start time of integration frame",
                "default_data": lambda: 1.234,
                'precision': '1 millisecond',
                'time_standard': 'UTC',
                'standard_name': 'time',
                'units': 'seconds since 2000-01-01 00:00:00.0',
            },
            "time_offset": {
                "format": "f4",
                "shape": ('unfiltered_events',),
                "fill_value": 9.96921e36,
                "long_name": "Time offset from epoch time",
                "default_data": lambda: np.linspace(0.0, 1000.0, nobs),
                'units': 'seconds',
            },
        }
    }


def l2_lef_schema(settings=None):
    """Define schema for LI L2 LEF product."""
    epoch_ts = datetime(2000, 1, 1, 0, 0, 0, 0)
    start_time = datetime.now()
    start_ts = (start_time - epoch_ts).total_seconds()

    settings = settings or {}

    nobs = settings.get("num_obs", 123)

    return {
        'providers': settings.get('providers', {}),
        'variable_path': settings.get('variable_path', 'data/'),
        'dimensions': {
            'events': nobs,
            'scalar': 1,
        },
        'variables': {
            "l1b_geolocation_warning": {
                "format": "i1",
                "shape": (),  # test explicitly the scalar case
                "default_data": lambda: 0
            },
            "l1b_missing_warning": {
                "format": "i1",
                "shape": ('scalar',),
                "default_data": lambda: 0
            },
            "l1b_radiometric_warning": {
                "format": "i1",
                "shape": ('scalar',),
                "default_data": lambda: 0
            },
        },
        'sector_variables': {
            "event_id": {
                "format": "u4",
                "shape": ('events',),
                "fill_value": 65535,
                "long_name": "ID of LI L2 Event",
                "default_data": lambda: np.arange(1, nobs + 1)
            },
            "group_id": {
                "format": "u4",
                "shape": ('events',),
                "fill_value": 65535,
                "long_name": "ID of associated LI L2 Group object",
                "default_data": lambda: np.arange(1, nobs + 1)
            },
            "flash_id": {
                "format": "u4",
                "shape": ('events',),
                "fill_value": 65535,
                "long_name": "ID of associated LI L2 Flash object",
                "default_data": lambda: np.arange(1, nobs + 1)
            },
            "detector": {
                "format": "u4",
                "shape": ('scalar',),
                "fill_value": 65535,
                "long_name": "ID of detector for this group",
                "default_data": lambda: 1
            },
            "latitude": {
                "format": "i2",
                "shape": ('events',),
                "fill_value": -32767,
                "long_name": "Latitude of group",
                "units": "degrees_north",
                "standard_name": "latitude",
                "default_data": lambda: np.linspace(-90, 90, nobs)
            },
            "longitude": {
                "format": "i2",
                "shape": ('events',),
                "fill_value": -32767,
                "long_name": "Longitude of group",
                "units": "degrees_east",
                "standard_name": "longitude",
                "default_data": lambda: np.linspace(-180, 80, nobs)
            },
            "radiance": {
                "format": "u2",
                "shape": ('events',),
                "long_name": "Radiance of Flash",
                "standard_name": "radiance",
                "units": "mW.m-2.sr-1",
                "default_data": lambda: np.clip(np.round(np.random.normal(500, 100, nobs)), 1, 2 ** 16 - 1)
            },
            "event_filter_qa": {
                "format": "u1",
                "shape": ('events',),
                "long_name": "L2 event pre-filtering quality assurance value",
                "default_data": lambda: np.random.randint(1, 2 ** 8 - 1, nobs)
            },
            "epoch_time": {
                "format": "f8",
                "shape": ('scalar',),
                "long_name": "Start time of integration frame",
                "units": "seconds since 2000-01-01 00:00:00.0",
                "default_data": lambda: start_ts
            },
            "time_offset": {
                "format": "f4",
                "shape": ('events',),
                "long_name": "Time offset from epoch time",
                "units": "seconds",
                "default_data": lambda: np.random.uniform(1, 2 ** 31 - 1, nobs)
            },
            "detector_row": {
                "format": "u2",
                "shape": ('events',),
                "long_name": "Detector row position of event pixel",
                "units": "1",
                "default_data": lambda: np.random.randint(1, 1000, nobs)
            },
            "detector_column": {
                "format": "u2",
                "shape": ('events',),
                "long_name": "Detector column position of event pixel",
                "units": "1",
                "default_data": lambda: np.random.randint(1, 1000, nobs)
            },
        }
    }


def l2_lgr_schema(settings=None):
    """Define schema for LI L2 LGR product."""
    settings = settings or {}
    ngrps = settings.get("num_groups", 120)

    return {
        'providers': settings.get('providers', {}),
        'variable_path': settings.get('variable_path', ''),
        'dimensions': {
            'groups': ngrps,
        },
        'variables': {
            "latitude": {
                "format": "f4",
                "shape": ('groups',),
                "long_name": "Latitude of group",
                "units": "degrees_north",
                "default_data": lambda: np.linspace(-90, 90, ngrps)
            },
            "longitude": {
                "format": "f4",
                "shape": ('groups',),
                "long_name": "Longitude of group",
                "units": "degrees_east",
                "default_data": lambda: np.linspace(-180, 80, ngrps)
            },
        }
    }


def l2_lfl_schema(settings=None):
    """Define schema for LI L2 LFL product."""
    settings = settings or {}

    nobs = settings.get("num_obs", 1234)
    epoch = datetime(2000, 1, 1)
    stime = (datetime(2019, 1, 1) - epoch).total_seconds()
    etime = (datetime(2019, 1, 2) - epoch).total_seconds()

    return {
        'providers': settings.get('providers', {}),
        'variable_path': settings.get('variable_path', ''),
        'dimensions': {
            'flashes': nobs,
        },
        'variables': {
            "latitude": {
                "format": "i2",
                "shape": ('flashes',),
                "long_name": "Latitude of Flash",
                "standard_name": "latitude",
                "units": "degrees_north",
                "add_offset": 0.0,
                "scale_factor": 0.0027,
                # Note: using a default range of [-88.3deg, 88.3deg] to stay in
                # the available type range [-32727,32727] with scaling:
                "default_data": lambda: np.linspace(-88.3 / 0.0027, 88.3 / 0.0027, nobs)
            },
            "longitude": {
                "format": "i2",
                "shape": ('flashes',),
                "long_name": "Longitude of Flash",
                "standard_name": "longitude",
                "units": "degrees_east",
                "add_offset": 0.0,
                "scale_factor": 0.0027,
                # Note: using a default range of [-88.3deg, 88.3deg] to stay in
                # the available type range [-32727,32727] with scaling:
                "default_data": lambda: np.linspace(-88.3 / 0.0027, 88.3 / 0.0027, nobs)
            },
            "radiance": {
                "format": "u2",
                "shape": ('flashes',),
                "long_name": "Radiance of Flash",
                "standard_name": "radiance",
                "units": "mW.m-2.sr-1",
                "default_data": lambda: np.round(np.random.normal(500, 100, nobs))
            },
            "flash_duration": {
                "format": "u2",
                "shape": ('flashes',),
                "long_name": "Flash duration",
                "standard_name": "flash_duration",
                "units": "ms",
                "default_data": lambda: np.linspace(0, 1000, nobs)
            },
            "flash_filter_confidence": {
                "format": "i1",
                "shape": ('flashes',),
                "long_name": "L2 filtered flash confidence",
                "standard_name": "flash_filter_confidence",
                "default_data": lambda: np.clip(np.round(np.random.normal(20, 10, nobs)), 1, 2 ** 7 - 1)
            },
            "flash_footprint": {
                "format": "u2",
                "shape": ('flashes',),
                "long_name": "Flash footprint size",
                "standard_name": 'flash_footprint',
                "units": "L1 grid pixels",
                "default_data": lambda: np.maximum(1, np.round(np.random.normal(5, 3, nobs)))
            },
            "flash_id": {
                "format": "u4",
                "shape": ('flashes',),
                "long_name": "Flash footprint size",
                "standard_name": 'flash_id',
                "default_data": lambda: np.arange(1, nobs + 1)
            },
            "flash_time": {
                "format": "f8",
                "shape": ('flashes',),
                "long_name": "Nominal flash time",
                "units": "seconds since 2000-01-01 00:00:00.0",
                "standard_name": 'time',
                "precision": "1 millisecond",
                "default_data": lambda: np.random.uniform(stime, etime, nobs)
            },
            "l1b_geolocation_warning": {
                "format": "u8",
                "shape": ('flashes',),
                "long_name": "L1b geolocation warning",
                "default_data": lambda: -127
            },
            "l1b_radiometric_warning": {
                "format": "u8",
                "shape": ('flashes',),
                "long_name": "L1b radiometric warning",
                "default_data": lambda: -127
            },
            "number_of_events": {
                "format": "u2",
                "shape": ('flashes',),
                "long_name": "Number of events in each flash",
                "default_data": lambda: 1
            },
            "number_of_groups": {
                "format": "u4",
                "shape": ('flashes',),
                "long_name": "Number of flashes in each flash",
                "default_data": lambda: 1
            },
        }
    }


def l2_af_schema(settings=None):
    """Define schema for LI L2 AF product."""
    settings = settings or {}
    nobs = settings.get("num_obs", 1234)

    return {
        'providers': settings.get('providers', {}),
        'variable_path': settings.get('variable_path', ''),
        'dimensions': accumulation_dimensions(1, nobs),
        'variables': {
            "accumulation_offsets": {
                "format": "u4",
                "shape": ('accumulations',),
                "default_data": lambda: 0
            },
            "accumulation_start_times": {
                "format": "f8",
                "shape": ('accumulations',),
                "default_data": lambda: 4.25055600161e8
            },
            "l1b_geolocation_warning": {
                "format": "i1",
                "shape": ('accumulations',),
                "long_name": "L1b geolocation warning",
                "default_data": lambda: -127
            },
            "l1b_radiometric_warning": {
                "format": "i1",
                "shape": ('accumulations',),
                "long_name": "L1b radiometric warning",
                "default_data": lambda: -127
            },
            "average_flash_qa": {
                "format": "i1",
                "shape": ('accumulations',),
                "default_data": lambda: 23
            },
            "flash_accumulation": {
                "format": "u2",
                "shape": ('pixels',),
                "default_data": lambda: np.clip(np.round(np.random.normal(1, 2, nobs)), 1, 2 ** 16 - 1)
            },
            "mtg_geos_projection": mtg_geos_projection(),
            "x": fci_grid_definition('X', nobs),
            "y": fci_grid_definition('Y', nobs),
        }
    }


def l2_afa_schema(settings=None):
    """Define schema for LI L2 AFA product."""
    settings = settings or {}
    npix = settings.get("num_pixels", 120)
    nacc = settings.get("num_accumulations", 20)

    return {
        'providers': settings.get('providers', {}),
        'variable_path': settings.get('variable_path', ''),
        'dimensions': accumulation_dimensions(nacc, npix),
        'variables': {
            "accumulation_start_times": {
                "format": "f4",
                "shape": ('accumulations',),
                "long_name": "Accumulation start time",
                "units": "seconds since 2000-01-01 00:00:00.0",
                "default_data": lambda: np.linspace(0.0, 1.0, nacc)
            },
            "accumulated_flash_area": {
                "format": "u4",
                "shape": ('pixels',),
                "fill_value": 4294967295,
                "long_name": "Number of contributing unique flashes to each pixel",
                "default_data": lambda: np.mod(np.arange(npix), 10) + 1
            },
            "mtg_geos_projection": mtg_geos_projection(),
            "x": fci_grid_definition('X', npix),
            "y": fci_grid_definition('Y', npix),
        }
    }


def l2_afr_schema(settings=None):
    """Define schema  for LI L2 AFR product."""
    settings = settings or {}
    nobs = settings.get("num_obs", 120)
    nacc = settings.get("num_accumulations", 20)

    return {
        'providers': settings.get('providers', {}),
        'variable_path': settings.get('variable_path', ''),
        'dimensions': accumulation_dimensions(nacc, nobs),
        'variables': {
            "flash_radiance": {
                "format": "f4",
                "shape": ('pixels',),
                "long_name": "Area averaged flash radiance accumulation",
                "grid_mapping": "mtg_geos_projection",
                "coordinate": "sparse: x y",
                "default_data": lambda: np.random.randint(low=1, high=6548, size=(120), dtype=np.int16)
            },
            "accumulation_start_times": {
                "format": "f4",
                "shape": ('accumulations',),
                "long_name": "Accumulation start time",
                "units": "seconds since 2000-01-01 00:00:00.0",
                "default_data": lambda: 0
            },
            "mtg_geos_projection": mtg_geos_projection(),
            "x": fci_grid_definition('X', nobs),
            "y": fci_grid_definition('Y', nobs),
        }
    }


def accumulation_dimensions(nacc, nobs):
    """Set dimensions for the accumulated products."""
    return {
        'accumulations': nacc,
        'pixels': nobs,
    }


def fci_grid_definition(axis, nobs):
    """FCI grid definition on X or Y axis."""
    if axis == 'X':
        long_name = 'azimuth angle encoded as column'
        standard_name = 'projection_x_coordinate'
    else:
        long_name = 'zenith angle encoded as row'
        standard_name = 'projection_y_coordinate'

    return {
        "format": "i2",
        "shape": ('pixels',),
        "add_offset": -0.155619516,
        "axis": axis,
        "long_name": long_name,
        "scale_factor": 5.58878e-5,
        "standard_name": standard_name,
        "units": 'radian',
        "valid_range": np.asarray([1, 5568]),
        "default_data": lambda: np.clip(np.round(np.random.normal(2000, 500, nobs)), 1, 2 ** 16 - 1)
    }


def mtg_geos_projection():
    """MTG geos projection definition."""
    return {
        "format": "i4",
        "shape": ('accumulations',),
        "grid_mapping_name": 'geostationary',
        "inverse_flattening": 298.2572221,
        "latitude_of_projection_origin": 0,
        "longitude_of_projection_origin": 0,
        "perspective_point_height": 42164000,
        "semi_major_axis": 6378169,
        "semi_minor_axis": 6356583.8,
        "sweep_angle_axis": 'y',
        "long_name": 'MTG geostationary projection',
        "default_data": lambda: -2147483647
    }


products_dict = {
    '2-LE': {'ftype': 'li_l2_le_nc', 'schema': l2_le_schema},
    '2-LEF': {'ftype': 'li_l2_lef_nc', 'schema': l2_lef_schema},
    '2-LGR': {'ftype': 'li_l2_lgr_nc', 'schema': l2_lgr_schema},
    '2-LFL': {'ftype': 'li_l2_lfl_nc', 'schema': l2_lfl_schema},
    '2-AF': {'ftype': 'li_l2_af_nc', 'schema': l2_af_schema},
    '2-AFA': {'ftype': 'li_l2_afa_nc', 'schema': l2_afa_schema},
    '2-AFR': {'ftype': 'li_l2_afr_nc', 'schema': l2_afr_schema},
}


def get_product_schema(pname, settings=None):
    """Retrieve an LI product schema given its name."""
    return products_dict[pname]['schema'](settings)


def extract_filetype_info(filetype_infos, filetype):
    """Extract Satpy-conform filetype_info from filetype_infos fixture."""
    ftype_info = filetype_infos[filetype]
    ftype_info['file_type'] = filetype
    return ftype_info


def set_variable_path(var_path, desc, sname):
    """Replace variable default path if applicable and ensure trailing separator."""
    vpath = desc.get('path', var_path)
    # Ensure we have a trailing separator:
    if vpath != "" and vpath[-1] != '/':
        vpath += '/'
    if sname != "":
        vpath += sname + "/"
    return vpath


def populate_dummy_data(data, names, details):
    """Populate variable with dummy data."""
    vname, sname = names
    desc, providers, settings = details
    if vname in providers:
        prov = providers[vname]
        # prov might be a function or directly an array that we assume will be of the correct shape:
        data[:] = prov(vname, sname, settings) if callable(prov) else prov
    else:
        # Otherwise we write the default data:
        if data.shape == ():
            # scalar case
            data = desc['default_data']()
        else:
            data[:] = desc['default_data']()


def add_attributes(attribs, ignored_attrs, desc):
    """Add all the custom properties directly as attributes."""
    for key, val in desc.items():
        if key not in ignored_attrs:
            attribs[key] = val


# Note: the helper class below has some missing abstract class implementation,
# but that is not critical to us, so ignoring them for now.
class FakeLIFileHandlerBase(FakeNetCDF4FileHandler):  # pylint: disable=abstract-method
    """Class for faking the NetCDF4 Filehandler."""

    # Optional parameter that may be provided at the time of the creation of this file handler
    # to customize the generated content. This may be either a simple dictionary or a callable
    # if a callable is provided it will be called to retrieve the actual parameter to be used:
    schema_parameters = None

    def get_variable_writer(self, dset, settings):
        """Get a variable writer."""
        # use a variable path prefix:
        var_path = settings.get("variable_path", "")

        # Also keep track of the potential providers:
        providers = settings.get('providers', {})

        # list of ignored attribute names:
        ignored_attrs = ["path", "format", "shape", "default_data", "fill_value"]

        # dictionary of dimensions:
        dims = settings.get('dimensions', {})

        def write_variable(vname, desc, sname=""):
            """Write a variable in our dataset."""
            # get numeric shape:
            shape_str = desc['shape']
            shape = tuple([dims[dname] for dname in shape_str])

            # Get the desired data type:
            dtype = TYPE_MAP[desc['format']]

            # Prepare a numpy array with the appropriate shape and type:
            data = np.zeros(shape, dtype=dtype)

            # Replace variable default path if applicable:
            vpath = set_variable_path(var_path, desc, sname)

            # Variable full name:
            full_name = f"{vpath}{vname}"

            # Add all the custom properties directly as attributes:
            attribs = {}
            add_attributes(attribs, ignored_attrs, desc)

            # Rename the fill value attribute:
            if 'fill_value' in desc:
                attribs['_FillValue'] = desc['fill_value']

            names = [vname, sname]
            details = [desc, providers, settings]
            populate_dummy_data(data, names, details)

            # Now we assign that data array:
            dset[full_name] = xr.DataArray(data, dims=shape_str, attrs=attribs)

            # Write the copy of the content:
            self.content[full_name] = data

        return write_variable

    def get_test_content(self, filename, filename_info, filetype_info):
        """Get the content of the test data.

        Here we generate the default content we want to provide depending
        on the provided filename infos.
        """
        # Retrieve the correct schema to write with potential customization parameters:
        params = FakeLIFileHandlerBase.schema_parameters
        if callable(params):
            # Note: params *IS* callable below:
            params = params(filename, filename_info, filetype_info)  # pylint: disable=not-callable

        settings = get_product_schema(filetype_info['file_desc']['product_type'], params)

        # Resulting dataset:
        dset = {}

        # Also keep a copy of the written content:
        self.content = {}

        # Retrieve the variable writer function
        write_variable = self.get_variable_writer(dset, settings)

        # Write all the raw (i.e not in sectors) variables:
        self.write_variables(settings, write_variable)

        # Write the sector variables:
        self.write_sector_variables(settings, write_variable)

        return dset

    def write_variables(self, settings, write_variable):
        """Write raw (i.e. not in sectors) variables."""
        if 'variables' in settings:
            variables = settings.get('variables')
            for vname, desc in variables.items():
                write_variable(vname, desc)

    def write_sector_variables(self, settings, write_variable):
        """Write the sector variables."""
        if 'sector_variables' in settings:
            sector_vars = settings.get('sector_variables')
            sectors = settings.get('sectors', ['north', 'east', 'south', 'west'])

            for sname in sectors:
                for vname, desc in sector_vars.items():
                    write_variable(vname, desc, sname)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2022 Satpy developers
#
# This file is part of satpy.
#
# satpy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# satpy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# satpy.  If not, see <http://www.gnu.org/licenses/>.
"""Geostationary High-speed Imager reader for the Level_1 HDF format.

This instrument is aboard the Fengyun-4B satellite. No document is available to describe this
format is available, but it's broadly similar to the co-flying AGRI instrument.

"""

import logging

from pyproj import Proj

from satpy.readers._geos_area import get_area_definition
from satpy.readers.fy4_base import FY4Base

logger = logging.getLogger(__name__)


class HDF_GHI_L1(FY4Base):
    """GHI l1 file handler."""

    def __init__(self, filename, filename_info, filetype_info):
        """Init filehandler."""
        super(HDF_GHI_L1, self).__init__(filename, filename_info, filetype_info)
        self.sensor = 'GHI'

    def get_dataset(self, dataset_id, ds_info):
        """Load a dataset."""
        ds_name = dataset_id['name']
        logger.debug('Reading in get_dataset %s.', ds_name)
        file_key = ds_info.get('file_key', ds_name)
        if self.CHANS_ID in file_key:
            file_key = f'Data/{file_key}'
        elif self.SUN_ID in file_key or self.SAT_ID in file_key:
            file_key = f'Navigation/{file_key}'
        data = self.get(file_key)
        if data.ndim >= 2:
            data = data.rename({data.dims[-2]: 'y', data.dims[-1]: 'x'})

        data = self.calibrate(data, ds_info, ds_name, file_key)

        self.adjust_attrs(data, ds_info)

        return data

    def adjust_attrs(self, data, ds_info):
        """Adjust the attrs of the data."""
        satname = self.PLATFORM_NAMES.get(self['/attr/Satellite Name'], self['/attr/Satellite Name'])
        data.attrs.update({'platform_name': satname,
                           'sensor': self['/attr/Sensor Identification Code'].lower(),
                           'orbital_parameters': {
                               'satellite_nominal_latitude': self['/attr/NOMSubSatLat'].item(),
                               'satellite_nominal_longitude': self['/attr/NOMSubSatLon'].item(),
                               'satellite_nominal_altitude': self['/attr/NOMSatHeight'].item()}})
        data.attrs.update(ds_info)
        # remove attributes that could be confusing later
        data.attrs.pop('FillValue', None)
        data.attrs.pop('Intercept', None)
        data.attrs.pop('Slope', None)

    def get_area_def(self, key):
        """Get the area definition."""
        # Coordination Group for Meteorological Satellites LRIT/HRIT Global Specification
        # https://www.cgms-info.org/documents/cgms-lrit-hrit-global-specification-(v2-8-of-30-oct-2013).pdf
        res = key['resolution']

        pdict = {}

        c_lats = self.file_content['/attr/Corner-Point Latitudes']
        c_lons = self.file_content['/attr/Corner-Point Longitudes']

        p1 = (c_lons[0], c_lats[0])
        p2 = (c_lons[1], c_lats[1])
        p3 = (c_lons[2], c_lats[2])
        p4 = (c_lons[3], c_lats[3])

        pdict['a'] = self.file_content['/attr/Semi_major_axis'] * 1E3  # equator radius (m)
        pdict['b'] = self.file_content['/attr/Semi_minor_axis'] * 1E3  # equator radius (m)
        pdict['h'] = self.file_content['/attr/NOMSatHeight'] * 1E3  # the altitude of satellite (m)

        pdict['h'] = pdict['h'] - pdict['a']

        pdict['ssp_lon'] = float(self.file_content['/attr/NOMSubSatLon'])
        pdict['nlines'] = float(self.file_content['/attr/RegLength'])
        pdict['ncols'] = float(self.file_content['/attr/RegWidth'])

        pdict['scandir'] = 'S2N'

        pdict['a_desc'] = "FY-4 {} area".format(self.filename_info['observation_type'])
        pdict['a_name'] = f'{self.filename_info["observation_type"]}_{res}m'
        pdict['p_id'] = f'FY-4, {res}m'

        proj_dict = {'a': pdict['a'],
                     'b': pdict['b'],
                     'lon_0': pdict['ssp_lon'],
                     'h': pdict['h'],
                     'proj': 'geos',
                     'units': 'm',
                     'sweep': 'y'}

        p = Proj(proj_dict)
        o1 = (p(p1[0], p1[1]))  # Upper left
        o2 = (p(p2[0], p2[1]))  # Upper right
        o3 = (p(p3[0], p3[1]))  # Lower left
        o4 = (p(p4[0], p4[1]))  # Lower right

        deller = res / 2.

        area = get_area_definition(pdict, (o3[0] - deller, o4[1] - deller, o2[0], o1[1]))

        return area
